unit Fixpunkt;

{$mode Delphi}

interface

uses
  SysUtils, Math;

type

  { TFixpointNumber }

  TFixpointNumber<E> = record
    Value: int64;
    class operator Implicit(const rhs: int64): TFixpointNumber<E>; inline;
    class operator Implicit(const rhs: double): TFixpointNumber<E>; inline;
    class operator Add(const lhs: TFixpointNumber<E>;
        const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    class operator Subtract(const lhs: TFixpointNumber<E>;
        const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    class operator Multiply(const lhs: TFixpointNumber<E>;
        const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    class operator Divide(const lhs: TFixpointNumber<E>;
        const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    class operator IntDivide(const lhs: TFixpointNumber<E>;
        const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    class operator Multiply(const lhs: TFixpointNumber<E>;
        const rhs: int64): TFixpointNumber<E>; inline;
    class operator Divide(const lhs: TFixpointNumber<E>;
        const rhs: int64): TFixpointNumber<E>; inline;
    class operator IntDivide(const lhs: TFixpointNumber<E>;
        const rhs: int64): TFixpointNumber<E>; inline;
    class operator Negative(const rhs: TFixpointNumber<E>): TFixpointNumber<E>; inline;
    function ToString(const Charset: TCharArray): string; overload;
    function ToString: string; overload;
    function ToDouble: double;
  end;

function numToString(Num: int64; const Base: integer;
  const CharSet: TCharArray): string; inline;

implementation

function numToString(Num: int64; const Base: integer; const CharSet: TCharArray): string;
var
  len, i: integer;
begin
  len := Ceil(logn(Base, num + 1));
  SetLength(Result, len);
  for i := len downto 1 do
  begin
    Result[i] := Charset[num mod Base];
    num := num div Base;
  end;
end;

{ TFixpointNumber }

class operator TFixpointNumber<E>.Implicit(const rhs: int64): TFixpointNumber<E>;
begin
  Result.Value := E.Divide(rhs);
end;

class operator TFixpointNumber<E>.Implicit(const rhs: double): TFixpointNumber<E>;
begin
  Result.Value := Round(E.Divide(rhs));
end;

class operator TFixpointNumber<E>.Add(const lhs: TFixpointNumber<E>;
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := lhs.Value + rhs.Value;
end;

class operator TFixpointNumber<E>.Subtract(const lhs: TFixpointNumber<E>;
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := lhs.Value - rhs.Value;
end;

class operator TFixpointNumber<E>.Multiply(const lhs: TFixpointNumber<E>;
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := E.Multiply(lhs.Value * rhs.Value);
end;

class operator TFixpointNumber<E>.Divide(const lhs: TFixpointNumber<E>;
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := E.Divide(lhs.Value) div rhs.Value;
end;

class operator TFixpointNumber<E>.IntDivide(const lhs: TFixpointNumber<E>;
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := E.Divide(lhs.Value) div rhs.Value;
end;

class operator TFixpointNumber<E>.Multiply(const lhs: TFixpointNumber<E>;
  const rhs: int64): TFixpointNumber<E>;
begin
  Result.Value := lhs.Value * rhs;
end;

class operator TFixpointNumber<E>.Divide(const lhs: TFixpointNumber<E>;
  const rhs: int64): TFixpointNumber<E>;
begin
  Result.Value := lhs.Value div rhs;
end;

class operator TFixpointNumber<E>.IntDivide(const lhs: TFixpointNumber<E>;
  const rhs: int64): TFixpointNumber<E>;
begin
  Result.Value := lhs.Value div rhs;
end;

class operator TFixpointNumber<E>.Negative(
  const rhs: TFixpointNumber<E>): TFixpointNumber<E>;
begin
  Result.Value := -rhs.Value;
end;

function TFixpointNumber<E>.ToString(const Charset: TCharArray): string;
var
  AbsVal: int64;
  IntVal: int64;
  IntPart: string;
  FracVal: int64;
  FracPart: string;
  len, i: integer;
begin
  if Value < 0 then
  begin
    AbsVal := -Value;
    Result := '-';
  end
  else
  begin
    AbsVal := Value;
    Result := '';
  end;
  if E.Exp < 0 then
  begin
    IntVal := E.Multiply(AbsVal);
    FracVal := Value - E.Divide(IntVal);
    IntPart := numToString(IntVal, E.Base, CharSet);
    FracPart := numToString(E.Divide(E.Base) + FracVal, E.Base, CharSet).Substring(2);
    Result += IntPart + '.' + FracPart;
  end
  else
  begin
    Result += numToString(AbsVal, E.Base, CharSet);
    len := Length(Result);
    SetLength(Result, len + E.Exp);
    for i := len to Length(Result) - 1 do
      Result[i + 1] := charset[0];
  end;
end;

function TFixpointNumber<E>.ToString: string;
var
  i: integer;
const
  charset: TCharArray = nil;
begin
  if charset = nil then
  begin
    Setlength(charset, E.Base);
    for i := 0 to E.Base - 1 do
    begin
      case i of
        0..9:
          charset[i] := chr(Ord('0') + i);
        10..35:
          charset[i] := chr(Ord('A') + i);
        36..61:
          charset[i] := chr(Ord('A') + i);
        else
          charset[i] := '#';
      end;
    end;
  end;
  Result := ToString(charset);
end;

function TFixpointNumber<E>.ToDouble: double;
begin
  Result := E.Multiply(Value.ToDouble);
end;

end.
